/*  FreeJ
 *  (c) Copyright 2001 Denis Roio aka jaromil <jaromil@dyne.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <iostream>
#include <assert.h>
#include <context.h>
#include <v4l.h>
#include <config.h>
#include "filters/filter-list.h"

#define W 640
#define H 480
#define D 16

/* this is the output context (screen) */
Context *screen = NULL;

/* this is a layer: actually only v4l is implemented */
V4lGrabber *grabber = NULL;

/* those are the pointers to filters 
   they have to be instantiated still */
Strobe *strobe = NULL;
Dumbcolor *dumbcolor = NULL;
Delaygrab *delaygrab = NULL;
Absdiff *absdiff = NULL;
Rotozoom *rotozoom = NULL;

int main (int argc, char **argv) {
  
  bool quit = false;
  bool show_fps = false;

  /* dirty hack to check filters are not added 2 times */
  bool actfilt[] = { false, false, false, false };

  fprintf(stderr,"%s version %s [ http://freej.dyne.org ]\n",PACKAGE,VERSION);
  cerr << "(c)2001 Denis Roio aka jaromil\n";
  SDL_Event event;

  /* allocates the context, SDL flags are used directly here */
  screen = new Context(W,H,D,SDL_HWSURFACE|SDL_DOUBLEBUF);
  if(screen->surf==NULL) exit(0);

  /* allocates v4l grabber layer */
  grabber = new V4lGrabber();
  if(!grabber->detect("/dev/video")) {
    delete grabber;
    delete screen;
    exit(0);
  }

  /* init the grabber */
  if(!grabber->init(screen,256,256,16)) {
    delete grabber;
    delete screen;
    exit(0);
  }
  /* center the layer */
  grabber->x = (Uint16)(W-256)/2;
  grabber->y = (Uint16)(H-256)/2;
  /* --------------------- */
  
  /* create delaygrab filter */
  delaygrab = new Delaygrab();
  /* create rotozoom filter */
  rotozoom = new Rotozoom();
  /* create dumbcolor filter */
  dumbcolor = new Dumbcolor();
  /* create strobo filter */
  strobe = new Strobe();
  /* create absolute difference filter */
  //  absdiff = new Absdiff();

  /* update the fps counter every 25 frames */
  screen->set_fps_interval(25);
  
  while(!quit) {
    while(SDL_PollEvent(&event)) {
      
      switch(event.type) {
      case SDL_QUIT:
	quit = true;
	break;
      case SDL_KEYDOWN:

	switch(event.key.keysym.sym) {
	  case SDLK_ESCAPE:
	  quit = true;
	  break;

	  /* add_filter adds the filters to the layer:
	     put the filter into the filter chain
	     gives to the filter informations about the layer
	     calls Filter.init() */
	case SDLK_1:
	  if(!actfilt[0]) {
	    grabber->add_filter(delaygrab);
	    actfilt[0] = true; }
	  break;
	case SDLK_2:
	  if(!actfilt[1]) {
	  grabber->add_filter(dumbcolor);
	    actfilt[1] = true; }
	  break;
	case SDLK_3:
	  if(!actfilt[2]) {
	  grabber->add_filter(rotozoom);
	    actfilt[2] = true; }
	  break;
	case SDLK_4:
	  if(!actfilt[3]) {
	  grabber->add_filter(strobe);
	  actfilt[3] = true; }
	  break;

	case SDLK_a:
	  delaygrab->createDelaymap(1);
	  break;

	case SDLK_b:
	  delaygrab->createDelaymap(2);
	  break;
	
	case SDLK_c:
	  delaygrab->createDelaymap(3);
	  break;
	case SDLK_f:
	  show_fps = !show_fps;
	  break;
	case SDLK_DELETE:
	  /* this on the next release will hopefully look better
	  if(absdiff!=NULL) {
	    delete absdiff;
	    absdiff = NULL; }
	  if(delaygrab!=NULL) {
	    delete delaygrab;
	    delaygrab = NULL; }
	  if(dumbcolor!=NULL) {
	    delete dumbcolor;
	    dumbcolor = NULL; }
	  if(rotozoom!=NULL) {
	    delete rotozoom;
	    rotozoom = NULL; }
	  if(strobe!=NULL) {
	    delete strobe;
	    strobe = NULL; } */
	  grabber->clear_filters();
	  for(int c=0;c<4;c++) actfilt[c] = false;
	  break;
	case SDLK_SPACE:
	  SDL_WM_ToggleFullScreen(screen->surf);
	  break;
	default:
	  break;
	  
	}
      }
    }

    /* blit the layer on screen */
    assert( grabber->cafudda() );

    /* flip the screen */
    assert( screen->flip() );

    if(show_fps)
      /* keep track of fps */
      if(!screen->calc_fps())
	/* and show them if counter says so (false)*/
	screen->print_fps();
    
    /* feed the layer (grabs new frames) */
    assert( grabber->feed() );
  }

  if(absdiff!=NULL)
    delete absdiff;
  if(delaygrab!=NULL)
    delete delaygrab;
  if(dumbcolor!=NULL)
    delete dumbcolor;
  if(rotozoom!=NULL)
    delete rotozoom;
  if(strobe!=NULL)
    delete strobe;
  if(grabber!=NULL)
    delete grabber;
  if(screen!=NULL)
    delete screen;
  exit(0);

}
