/*  FreeJ
 *  (c) Copyright 2001 Denis Roio aka jaromil <jaromil@dyne.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <iostream>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include <v4l.h>

#define PALETTE_8 VIDEO_PALETTE_GREY
#define PALETTE_16 VIDEO_PALETTE_RGB565
#define PALETTE_24 VIDEO_PALETTE_RGB24
#define PALETTE_32 VIDEO_PALETTE_RGB32

V4lGrabber::V4lGrabber() {
  dev = -1;
  input = 1;
  buffer = NULL;
}

V4lGrabber::~V4lGrabber() {
  cerr << "-- V4lGrabber::~V4lGrabber\n";
  if(device!=NULL) {
    fprintf(stderr,"   closing device[%s] ", device);
    free(device);
  }
  if(dev>0) {
    fprintf(stderr," fd[%u]\n",dev);
    close(dev);
  } else cerr << endl;
  if(buffer!=NULL) {
    fprintf(stderr,
	    "   unmapping buffer[0x%lx] size[%u]\n",(long unsigned int)buffer,grab_map.size);
    munmap(buffer,grab_map.size);
  }
}

bool V4lGrabber::detect(char *devfile) {
  int counter, res;
  char *capabilities[] = {
    "     VID_TYPE_CAPTURE          can capture to memory",
    "     VID_TYPE_TUNER            has a tuner of some form",
    "     VID_TYPE_TELETEXT         has teletext capability",
    "     VID_TYPE_OVERLAY          can overlay its image onto the frame buffer",
    "     VID_TYPE_CHROMAKEY        overlay is chromakeyed",
    "     VID_TYPE_CLIPPING         overlay clipping supported",
    "     VID_TYPE_FRAMERAM         overlay overwrites frame buffer memory",
    "     VID_TYPE_SCALES           supports image scaling",
    "     VID_TYPE_MONOCHROME       image capture is grey scale only",
    "     VID_TYPE_SUBCAPTURE       capture can be of only part of the image"
  };

  device = strdup(devfile);

  if (-1 == (dev = open(device,O_RDWR))) {
    perror(" ! Error in opening video capture device");
    return(false);
  }
  
  res = ioctl(dev,VIDIOCGCAP,&grab_cap);
  if(res<0) {
    perror(" ! Error in VIDIOCGCAP ");
    return(false);
  }
  
  fprintf(stderr,"-- V4lGrabber::detect\n");
  fprintf(stderr,"   Device used is %s\n",device);
  fprintf(stderr,"   %s\n",grab_cap.name);
  fprintf(stderr,"   %u channels detected\n",grab_cap.channels);
  fprintf(stderr,"   max size w[%u] h[%u] - min size w[%u] h[%u]\n",grab_cap.maxwidth,grab_cap.maxheight,grab_cap.minwidth,grab_cap.minheight);
  fprintf(stderr,"   Video capabilities:\n");
  for (counter=0;counter<11;counter++)
    if (grab_cap.type & (1 << counter)) fprintf(stderr,"%s\n",capabilities[counter]);
  if((grab_pic.palette|VIDEO_PALETTE_GREY)&VIDEO_PALETTE_GREY)
    fprintf(stderr,"     VIDEO_PALETTE_GREY        device is able to grab greyscale frames\n");
  else {
    grab24 = 1;
  }

  if(strncmp(grab_cap.name,"Buz",3)==0)
    /* the device is a iomega Buz. We need to check that 'cause the iomegaBuz
       claims to have the VIDEO_PALETTE_GREY capability, but then grabs 24bit
       images, overflowing the buffer. So we don't trust the capability report
       and turn on software 24bit->8bit downsampling */
    grab24 = 1;

  if(strncmp(grab_cap.name,"OV511",5)==0)
    /* the device is a USB camera, chipset OV511, wich has only one input
       channel (afaik, as long as i do not have one) so we force that chan */
    input = 0;
  
  if((grab_cap.type|VID_TYPE_TUNER)&VID_TYPE_TUNER)
    /* the device does'nt has any tuner, so we avoid some ioctl
       this should be a fix for Creative Webcam II. thanks to Ben Wilson */
    have_tuner = 1;
  
  /* set the minwidth and minheight */
  minw = grab_cap.minwidth;
  minh = grab_cap.minheight;

  if (ioctl (dev, VIDIOCGMBUF, &grab_map) == -1) {
    perror (" ! Error in ioctl VIDIOCGMBUF");
    return(false);
  }
  /* print memory info */
  fprintf (stderr,"   Memory Map of %i frames: %i bytes\n",grab_map.frames,grab_map.size);
  for(counter=0;counter<grab_map.frames;counter++)
    fprintf (stderr,"     Offset of frame %i: %i\n",counter,grab_map.offsets[counter]);
  num_frame = grab_map.frames;
  cerr << "   grabber succesfully detected\n";
  return(true);
}

/* v4l palettes 
   VIDEO_PALETTE_GREY     b/w
   VIDEO_PALETTE_HI240    8bit
   VIDEO_PALETTE_RGB565   16bit
   VIDEO_PALETTE_RGB24    24bit
   VIDEO_PALETTE_YUV422   YUV8bit (4Y 2U 2V)
   others... */

bool V4lGrabber::init(Context *screen,int wdt, int hgt, int dpt) {
  int i;
  cerr << "-- V4lGrabber::init\n";

  /* set image source and TV norm */
  grab_chan.channel = input;
  
  if(have_tuner) {
    /* does this only if the device has a tuner */
    if (-1 == ioctl(dev,VIDIOCGCHAN,&grab_chan)) {
      perror(" ! Error in ioctl VIDIOCGCHAN ");
      return(false);
    }
    
    if (-1 == ioctl(dev,VIDIOCSCHAN,&grab_chan)) {
      perror(" ! Error in ioctl VIDIOCSCHAN ");
      return(false);
    }
  }
  
  if (-1 == ioctl(dev, VIDIOCGPICT, &grab_pic)) {
    perror(" ! Error ioctl VIDIOCGPICT ");
    return(false);
  }

  /* v4lgrabber can work only with the screen size now
     TODO: check with minwidth maxwidth */

  if(wdt==0) w = screen->w; else w = wdt;
  if(hgt==0) h = screen->h; else h = hgt;
  if(dpt==0) bpp = screen->bpp; else bpp = dpt;
  size = w*h*(bpp>>3);
  pitch = w*(bpp>>3);
  fps = &screen->fps;
  
  switch(bpp) {
  case 8:
    palette = PALETTE_8;
  case 16:
    palette = PALETTE_16;
    break;
  case 24:
    palette = PALETTE_24;
    break;
  case 32:
    palette = PALETTE_32;
    break;
  default:
    fprintf(stderr," ! V4lGrabber does not support Contexts with bpp[%u]\n",bpp);
    return(false);
    break;
  }
  
  for(i=0; i<grab_map.frames; i++) {
    grab_buf[i].format = palette;
    grab_buf[i].frame  = i;
    grab_buf[i].height = h;
    grab_buf[i].width = w;
  }
  
  cur_frame = ok_frame = 0;
  
  /* mmap (POSIX.4) buffer for grabber device */
  buffer = (unsigned char *) mmap(0,grab_map.size,PROT_READ|PROT_WRITE,MAP_SHARED,dev,0);
  if(MAP_FAILED == buffer) {
    perror(" ! Cannot allocate v4lgrabber buffer ");
    return(false);
  }
  
  /* grab the first frames */
  for(i=0;i<grab_map.frames;i++) {
    if (-1 == ioctl(dev,VIDIOCMCAPTURE,&grab_buf[i])) {
      perror(" ! V4lGrabber::init ioctl VIDIOCMCAPTURE");
      return(false);
    }
  }

  /* INIT from the LAYER CLASS */
  _init();

  this->screen = screen;

  fprintf(stderr,
	  "   new layer registered :: w[%u] h[%u] bpp[%u] size[%u] grab_mmap[%u]\n"
	  ,w,h,bpp,size,size*num_frame);
  
  return(true);
}

unsigned char *V4lGrabber::get_buffer() {
  return(&buffer[grab_map.offsets[ok_frame]]);
}

bool V4lGrabber::feed() {

  ok_frame = cur_frame;
  cur_frame = ((cur_frame+1)%num_frame);
  grab_buf[0].format = palette;
  
  if (-1 == ioctl(dev,VIDIOCSYNC,&grab_buf[cur_frame]))
    perror("-- V4lGrabber::feed\n"
	   " ! Error in ioctl VIDIOCSYNC");
  
  if (-1 == ioctl(dev,VIDIOCMCAPTURE,&grab_buf[cur_frame])) 
    perror("-- V4lGrabber::feed\n"
	   " ! Error in ioctl VIDIOCMCAPTURE");

  return(1);
  
}
